/*
 * libkarma/karma.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "karma.h"
#include "properties.h"
#include "md5.h"

static void cat_int32(char * dst, uint32_t cat);
static void cat_int64(char * dst, uint64_t cat);
static int busy_handler(int rio, uint32_t waitfor);

static uint32_t recv_header(int rio);
static int32_t recv_status(int rio);
static int32_t recv_properties(int rio, uint32_t identifier, uint32_t r_id, char ** properties);

/*
	 returns a socket, which is connect'ed to the rio karma or -1
*/
int connectkarma(char * ip)
{
	struct hostent *host;
	struct sockaddr_in dest;
	int rio;
	
	host = gethostbyname(ip);
	if(host == NULL)
		return -1;

	rio = socket(AF_INET, SOCK_STREAM, 0);
	if(rio == -1)
		return -1;
	
	memset(&dest, 0, sizeof(dest));
	dest.sin_addr = *(struct in_addr *)host->h_addr_list[0];
	dest.sin_family = AF_INET;
	dest.sin_port = htons(RIOPORT);

	connect(rio, (struct sockaddr *)&dest, sizeof(dest));

	return rio;	
}


static void cat_int32(char * dst, uint32_t cat)
{
	*dst=cat % 256;
	*++dst=(cat >>  8) % 256;
	*++dst=(cat >> 16) % 256;
	*++dst=(cat >> 24) % 256;
}

static void cat_int64(char * dst, uint64_t cat)
{
	*dst=cat % 256;
	*++dst=(cat >>  8) % 256;
	*++dst=(cat >> 16) % 256;
	*++dst=(cat >> 24) % 256;
	*++dst=(cat >> 32) % 256;
	*++dst=(cat >> 40) % 256;
	*++dst=(cat >> 48) % 256;
	*++dst=(cat >> 56) % 256;
}

/*
	 {thanks Cliff L. Biffle...}
	 computes the RID of a file
	 rid algorithm pseudocode:
	 if(length of file <= 64k) 
	 	return MD5(file);
	 else
	 	A = MD5(first 64k);
	 	B = MD5(last 64k);
	 	C = MD5(64k starting at (length - 64k)/2);
	 	return A ^ B ^ C; 
	 
	 oggs,flacs: offset=0 length=length of file
	 mp3s: offset=after the tags length=length of file -128
*/
#define RID_BLOCKS 65536
char * generate_rid(int file, int offset, int length)
{
	static unsigned char rid_output[32];
	int i;
	uint32_t size;
	char cache[RID_BLOCKS];
	unsigned char md5output[16], b_md5[16], c_md5[16];
	char *ret;
	md5_context ctx;
	
	if(lseek(file, offset, SEEK_SET)==-1) return NULL;
	size=read(file, &cache, RID_BLOCKS);
	
	md5_starts(&ctx);
	if(length-offset < RID_BLOCKS){
		md5_update(&ctx, (uint8 *)&cache, size);
		md5_finish(&ctx, md5output);
	}else{
		md5_update(&ctx, cache, RID_BLOCKS);
		md5_finish(&ctx, md5output);
		lseek(file, length-RID_BLOCKS, SEEK_SET);
		read(file, &cache, RID_BLOCKS);
		md5_starts(&ctx);
		md5_update(&ctx, cache, RID_BLOCKS);
		md5_finish(&ctx, b_md5);
		lseek(file, (offset + length - RID_BLOCKS)/2, SEEK_SET);
		size=read(file, &cache, RID_BLOCKS);
		md5_starts(&ctx);
		md5_update(&ctx, cache, size);
		md5_finish(&ctx, c_md5);
		for(i=0; i < 16; i++)
			md5output[i]=md5output[i]^b_md5[i]^c_md5[i];
	}
	for(i=0; i<16;i++){
		if((md5output[i]>>4) < 10) rid_output[i*2]=(md5output[i]>>4)+'0';
		else rid_output[i*2]=(md5output[i]>>4)+87;
		if((md5output[i]%16) < 10) rid_output[i*2+1]=(md5output[i]%16)+'0';
		else rid_output[i*2+1]=(md5output[i]%16)+87;
	}
	rid_output[32]='\0';
	ret=&rid_output[0];
	return ret;
}


int send_request(int rio, uint32_t identifier, char * payload, int plen)
{
	char msg[HEADERLENGTH];

	/*protocol signature*/
	cat_int32(msg, RIO_HEADER);
	/*command number from identifier*/
	cat_int32(&msg[4],identifier);

	if(payload==NULL)
		send(rio, msg, HEADERLENGTH, MSG_FIN);
	else{
		send(rio, msg, HEADERLENGTH, 0);
		send(rio, payload, plen, MSG_FIN);
	}

	return 0;
}

/*
	 recv the header of the karma,
	 it'll return -1 if the first part of the header
	 doesn't match with the standard rio header, but if it
	 does, it'll return the command identifier
*/
static uint32_t recv_header(int rio)
{
	uint32_t h;
	
	recv(rio, &h, sizeof(h), MSG_WAITALL);
	if(h!=RIO_HEADER)
		return -1;
	recv(rio, &h, sizeof(h), MSG_WAITALL);
	return h;
}

/*
	 returns the status as int32_t
*/
static int32_t recv_status(int rio)
{
	int32_t s;
	recv(rio, &s, sizeof(s), MSG_WAITALL);
	return s;
}

/*
	 recv the properties,
	 copies the data to **properties, 
	 memory will be allocated by recv_properties,
	 returns -1 if the header doesn't match with identifier 
	 or the status indicates an failure.
*/
static int32_t recv_properties(int rio, uint32_t identifier, uint32_t r_id, char ** properties)
{
	uint64_t i=0;
	char id[3];

	if(r_id > 0){
		cat_int32(id, r_id);
		send_request(rio, identifier, id, 4);
	}else	send_request(rio, identifier, NULL, 0);

	*properties=NULL;
	
	if(busy_handler(rio, identifier)){
		if(recv_status(rio) != 0) 
			return -1;
	}else return -1;

	do{
		if(i%8192==0)
			*properties=(char *)realloc(*properties, (i+8193));
		(*properties)[i+4]='!';
		recv(rio, &(*properties)[i], 4, MSG_WAITALL);
		i+=4;
	}while((*properties)[i-1]!=0);

	return 0;
}

static int busy_handler(int rio, uint32_t waitfor)
{
	uint32_t header;
	header=recv_header(rio);
	if(header==BUSY){
		do{
			recv_status(rio);
			recv_status(rio);
		}while((header=recv_header(rio))==BUSY);
	}
	return header==waitfor;
}

/*
	 writes the protocol version to *major_version and *minor_version, 
	 returns -1 if the header doesn't  match with GET_PROTOCOL_VERSION
 */
int get_protocol_version(int rio, uint32_t * major_version, uint32_t * minor_version)
{
	char versions[8];

	cat_int32(versions, *major_version);
	cat_int32(&versions[4], *minor_version);

	send_request(rio, 0, versions, 8);
	if(!busy_handler(rio, GET_PROTOCOL_VERSION))
		return -1;

	recv(rio, major_version, sizeof(uint32_t), 0);
	recv(rio, minor_version, sizeof(uint32_t), 0);
	return 0;
}


int32_t get_authentication_salt(int rio, char ** salt)
{
	*salt=(char *)malloc(SALTLENGTH);

	send_request(rio, GET_AUTHENTICATION_SALT, NULL, 0);
	if(!busy_handler(rio, GET_AUTHENTICATION_SALT)) 
		return -1;
	recv(rio, *salt, SALTLENGTH, 0);

	return 0;
}


int32_t authenticate(int rio, char * pass)
{
	md5_context ctx;
	char * md5input;
	char md5sum[16];

	char * salt;
	uint32_t rights;


	if(get_authentication_salt(rio, &salt) == -1)
		return -1;

	md5input=(char *)malloc((strlen(pass)+SALTLENGTH+1));
	memcpy(md5input, salt, SALTLENGTH);
	memcpy(&md5input[SALTLENGTH], pass, strlen(pass));

	md5_starts(&ctx);
	md5_update(&ctx, (uint8 *)md5input, SALTLENGTH+strlen(pass));
	md5_finish(&ctx, md5sum);

	free(salt);
	free(md5input);

	send_request(rio, AUTHENTICATE, md5sum, 16);
	if(!busy_handler(rio, AUTHENTICATE)) return -1;
	recv_status(rio); /*we don't care about the status, because we want to get the rights...*/
	recv(rio, &rights, sizeof(rights), 0);
	return rights;
}


int get_device_details(int rio, char **name, char **version, uint32_t * storagedevices)
{
	*name=(char *)malloc(DETAILSLENGTH);
	*version=(char *)malloc(DETAILSLENGTH);

	send_request(rio, GET_DEVICE_DETAILS, NULL, 0);
	if(!busy_handler(rio, GET_DEVICE_DETAILS)) return -1;

	recv(rio, *name, DETAILSLENGTH, 0);
	recv(rio, *version, DETAILSLENGTH, 0);
	recv(rio, storagedevices, sizeof(uint32_t), 0);
	return 0;
}


int get_storage_details(int rio, uint32_t storage_id, uint32_t *n_files, 
		uint64_t *s_size, uint64_t *f_space, uint32_t * highest_file_id)
{
	char id[3];

	cat_int32(id, storage_id);
	send_request(rio, GET_STORAGE_DETAILS, id, sizeof(uint32_t));

	if(!busy_handler(rio, GET_STORAGE_DETAILS)) return -1;
	if(recv_status(rio) != 0) return -1;

	recv(rio, n_files, sizeof(uint32_t), 0);
	recv(rio, s_size, sizeof(uint64_t), 0);
	recv(rio, f_space, sizeof(uint64_t), 0);
	recv(rio, highest_file_id, sizeof(uint32_t), 0);

	return 0;
}

int get_device_settings(int rio, char ** properties)
{
	return recv_properties(rio, GET_DEVICE_SETTINGS, 0, properties);
}


int32_t update_device_settings(int rio, char * properties)
{
	int len;
	len=strlen(properties);
	while(len%4) ++len;
	send_request(rio, UPDATE_DEVICE_SETTINGS, properties, len);
	if(busy_handler(rio, UPDATE_DEVICE_SETTINGS))
		return recv_status(rio);
	else
		return -1;
}

int32_t request_io_lock(int rio, uint32_t type)
{
	char p[3];

	cat_int32(p, type);
	send_request(rio, REQUEST_IO_LOCK, p, 4);
	if(busy_handler(rio, REQUEST_IO_LOCK))
		return recv_status(rio);
	else
		return -1;
}

int32_t release_io_lock(int rio)
{
	send_request(rio, RELEASE_IO_LOCK, NULL, 0);
	if(busy_handler(rio, RELEASE_IO_LOCK))
		return recv_status(rio);
	else 
		return -1;
}

int32_t prepare(int rio, uint64_t size, uint32_t * file_id, uint32_t storage_id)
{
	char p[15];

	cat_int64(p, size);
	cat_int32(&p[4], *file_id);
	cat_int32(&p[12], storage_id);

	send_request(rio, PREPARE, p, 16);
	if(!busy_handler(rio, PREPARE))
		return -1;
	recv(rio, file_id, sizeof(file_id), 0);
	return recv_status(rio);
}

int32_t write_file_chunk(int rio, uint64_t offset, uint64_t size, uint32_t file_id,
		    uint32_t storage_id, char * data)
{
	char * packet;

	packet=malloc(sizeof(uint64_t)*3+size);
	cat_int64(packet, offset);
	cat_int64(&packet[8], size);
	cat_int32(&packet[16], file_id);
	cat_int32(&packet[20], storage_id);
	memcpy(&packet[24], data, size);

	send_request(rio, WRITE_FILE_CHUNK, packet, size+sizeof(uint64_t)*3);
	free(packet);
	
	if(busy_handler(rio, WRITE_FILE_CHUNK)) 
		return recv_status(rio);
	else 
		return -1;
}

int32_t get_all_file_details(int rio, char ** properties)
{
	return recv_properties(rio, GET_ALL_FILE_DETAILS, 0, properties);
}

int32_t get_file_details(int rio, uint32_t file_id, char ** properties)
{
	return recv_properties(rio, GET_FILE_DETAILS, file_id, properties);
}

int32_t update_file_details(int rio, uint32_t file_id, char * properties)
{
	char * packet;
	int len,rlen;

	rlen=strlen(properties);
	len=rlen;
	++len;
	while(len%4) ++len;
	packet=malloc(sizeof(uint32_t)+len);
	cat_int32(packet, file_id);
	memcpy(&packet[4], properties, len);
	memset(&packet[rlen+sizeof(uint32_t)], 0, len-rlen);
	
	send_request(rio, UPDATE_FILE_DETAILS, packet, len+sizeof(uint32_t));
	free(packet);
	
	if(busy_handler(rio, UPDATE_FILE_DETAILS))
		return recv_status(rio);
	else
		return -1;
}

int32_t read_file_chunk(int rio, uint64_t offset, uint64_t size, uint32_t file_id,
		    char ** data, uint64_t * retsize)
{
	char p[20];
	uint64_t realsize;
	
	*retsize=0;
	cat_int64(p, offset);
	cat_int64(&p[8], size);
	cat_int32(&p[16], file_id);
	send_request(rio, READ_FILE_CHUNK, p, 20);

	if(!busy_handler(rio, READ_FILE_CHUNK)) return -1;
	recv(rio, &realsize, sizeof(uint64_t), 0);
	*data=(char *)malloc(realsize+1);
	if(recv_status(rio) != 0) return -1;
	*retsize=recv(rio, *data, realsize, MSG_WAITALL);

	return recv_status(rio);
}
	

int32_t delete_file(int rio,uint32_t file_id)
{
	char id[3];

	cat_int32(id, file_id);
	send_request(rio, DELETE_FILE, id, 4);
	if(busy_handler(rio, DELETE_FILE))
		return recv_status(rio);
	else 
		return -1;
}

/*
	 Now, that command would be scary ;)
	 (It's not supported by the rio karma ;))
*/
int32_t format_storage(int rio, uint32_t storage_id)
{
	char id[3];

	cat_int32(id, storage_id);
	send_request(rio, FORMAT_STORAGE, id, 4);
	if(busy_handler(rio, FORMAT_STORAGE))
		return recv_status(rio);
	else
		return -1;
}

int32_t hangup(int rio)
{
	send_request(rio, HANGUP, NULL, 0);
	if(busy_handler(rio, HANGUP))
		return recv_status(rio);
	else
		return -1;
}

int32_t device_operation(int rio, uint64_t size, char * data, char ** newdata)
{
	char * d;
	int32_t status;
	
	d=malloc(sizeof(char)*sizeof(uint64_t)+sizeof(char)*size);

	cat_int64(d, size);
	memcpy(&d[8], data, size);
	send_request(rio, DEVICE_OPERATION, d, size+sizeof(uint64_t));

	if(!busy_handler(rio, DEVICE_OPERATION)) return -1;

	recv(rio, &size, sizeof(uint64_t), 0);
	status=recv_status(rio);
	if(status != 0) return status;
	*newdata=realloc(*newdata, size);
	recv(rio, *newdata, size, MSG_WAITALL);
	
	return status;	
}
